package ostate;

import ostate.util.tree.StatesTree;
import ostate.util.tree.Node;
import ostate.util.properties.Properties;
import ostate.discriptions.StateDiscriptor;
import ostate.discriptions.StateDiscriptorFactory;
import ostate.discriptions.ANDStateDiscriptor;
import ostate.discriptions.ORStateDiscriptor;
import ostate.exceptions.StateDoesNotImplementException;
import ostate.exceptions.InvalidStatesConfigurationException;
import ostate.exceptions.InternalErrorException;
import ostate.invoke.impl.DefaultMethodInvoker;
import ostate.invoke.MethodInvoker;
import ostate.invoke.EventProcessor;
import ostate.invoke.impl.MethodInvokingContextImpl;
import ostate.invoke.impl.EventProcessingContextImpl;

import java.util.List;
import java.util.ArrayList;
import java.util.Stack;
import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;

/**
 * Class, used to perform actions on a {@link StatesTree}.
 *
 * @author Pavel Mavrin
 * @version $Id$
 */

class StatesTreeController<I> {

    private final StatesTree<I> statesTree;

    private final IStateFactory<I> stateFactory;

    private final Class<I> AI;

    private final Stack<IState<I>> sources;

    public StatesTreeController(StatesTree<I> statesTree, IStateFactory<I> stateFactory, Class<I> AI, Stack<IState<I>> sources) {
        this.statesTree = statesTree;
        this.stateFactory = stateFactory;
        this.AI = AI;
        this.sources = sources;
    }

    public void processTransition(Transition<I> transition) {
        if (OStateProperties.isDebug()) {
            String s1 = transition.getSource() == null ?
                    "null" :
                    StateDiscriptorFactory.getStateDiscriptor(transition.getSource()).getName();
            String s2 = StateDiscriptorFactory.getStateDiscriptor(transition.getTarget()).getName();
            System.out.println("Transition: " + s1 + " -> " + s2);
        }
        IState<I> source = transition.getSource();
        if (source != null) {
            Node<IState<I>> sourceNode = statesTree.getNode(source.getClass());
            if (sourceNode == null || !sourceNode.getValue().equals(source)) {
                // Source state is not active
                return;
            }
        }

        Class<? extends IState<I>> target = transition.getTarget();
        Properties properties = transition.getProperties();

        if (statesTree.getNode(target) != null) {
            // Target state is active
            Node<IState<I>> node = statesTree.getNode(target);
            invokeMethodSafe(
                    node,
                    StateMachineManager.STATE_STOP_METHOD,
                    DefaultMethodInvoker.CHILDREN_ROOT);
            node.removeAllChildren();
            node.setValue(constructState(target, properties));
            fillBranch(statesTree, node, properties);
            return;
        }

        List<IState<I>> newStates = new ArrayList<IState<I>>();

        Class<? extends IState<I>> stateClass = target;

        //Find the nearest active parent
        while (stateClass != null && statesTree.getNode(stateClass) == null) {
            IState<I> newState = constructState(
                    stateClass,
                    properties);
            newStates.add(newState);
            stateClass = StateDiscriptorFactory.getStateDiscriptor(newState).getParent();
            if (stateClass.equals(IState.class)) {
                stateClass = null;
            }
        }

        Node<IState<I>> lastNode;
        if (stateClass == null) {
            //New root state, remove all states
            lastNode = null;
            invokeMethodSafe(
                    statesTree.getRoot(),
                    StateMachineManager.STATE_STOP_METHOD,
                    DefaultMethodInvoker.CHILDREN_ROOT);
        } else {
            // Remove branch
            lastNode = statesTree.getNode(stateClass);
            if (lastNode.getChildren().size() != 1) {
                throw new InvalidStatesConfigurationException();
            }
            invokeMethodSafe(
                    lastNode.getChildren().get(0),
                    StateMachineManager.STATE_STOP_METHOD,
                    DefaultMethodInvoker.CHILDREN_ROOT);
            lastNode.removeAllChildren();
        }

        //Run transition action
        if (transition.getAction() != null) {
            sources.push(source);
            transition.getAction().run();
            sources.pop();
        }

        if (target == null) {
            return;
        }

        Node<IState<I>> rootNode = null;

        for (int i = newStates.size() - 1; i >= 0; i--) {
            IState<I> state = newStates.get(i);
            if (lastNode == null) {
                statesTree.setRoot(state);
                lastNode = statesTree.getRoot();
            } else {
                Node<IState<I>> newNode = new Node<IState<I>>(
                        statesTree,
                        lastNode,
                        state);
                addChildNodeTo(newNode, lastNode);
                lastNode = newNode;
            }
            if (StateDiscriptorFactory.getStateDiscriptor(state) instanceof ANDStateDiscriptor) {
                ANDStateDiscriptor<I> newStateDiscriptor =
                        (ANDStateDiscriptor<I>) StateDiscriptorFactory.getStateDiscriptor(state);
                for (Class<? extends IState<I>> partClass : newStateDiscriptor.getParts()) {
                    if (i > 0 && !partClass.equals(newStates.get(i - 1).getClass())) {
                        addBranch(statesTree, partClass, lastNode, properties);
                    }
                }
            }
            if (i == newStates.size() - 1) {
                rootNode = lastNode;
            }
        }
        fillBranch(statesTree, lastNode, properties);

        if (OStateProperties.isDebug())
            printTree(statesTree.getRoot(), 0);

        invokeMethodSafe(
                rootNode,
                StateMachineManager.STATE_START_METHOD,
                DefaultMethodInvoker.ROOT_CHILDREN);
    }

    private void printTree(Node<IState<I>> node, int d) {
        String s = "";
        for (int i = 0; i < d; i++) s = s + " ";
        System.out.println(s + StateDiscriptorFactory.getStateDiscriptor(node.getValue()).getName());
        for (Node<IState<I>> child : node.getChildren()) {
            printTree(child, d + 1);
        }
    }

    public Object invokeMethod(Node<IState<I>> node, Method method, MethodInvoker invoker, Object... args)
            throws IllegalAccessException, InvocationTargetException {
        if (node != null) {
            return invoker.invoke(
                    new MethodInvokingContextImpl<I>(
                            node,
                            method,
                            invoker,
                            sources,
                            args)).getResult();

        } else {
            return null;
        }
    }

    public void processEvent(Node<IState<I>> node, Event event, EventProcessor processor) {
        try {
            processor.processEvent(
                    new EventProcessingContextImpl<I>(
                            statesTree.getRoot(),
                            event,
                            processor,
                            sources));
        } catch(IllegalAccessException e) {
            throw new InternalErrorException("Cannot access handler");
        } catch(InvocationTargetException e) {
            Throwable throwable = e.getTargetException();
            if (throwable instanceof Error) {
                throw (Error)throwable;
            } else if (throwable instanceof RuntimeException) {
                throw (RuntimeException)throwable;
            } else {
                throw new InternalErrorException("Unexpected excepton", throwable);
            }
        }
    }

    private Object invokeMethodSafe(Node<IState<I>> node, Method method, MethodInvoker invoker, Object... args) {
        try {
            return invokeMethod(node, method, invoker, args);
        } catch(IllegalAccessException e) {
            throw new IllegalStateException("Method is not public: " + method);
        } catch(InvocationTargetException e) {
            Throwable throwable = e.getTargetException();
            if (throwable instanceof Error) {
                throw (Error)throwable;
            } else if (throwable instanceof RuntimeException) {
                throw (RuntimeException)throwable;
            } else {
                throw new InternalErrorException("Unexpected excepton", throwable);
            }
        }
    }

    private IState<I> constructState(Class<? extends IState<I>> stateClass, Properties properties) {
        IState<I> state = stateFactory.constructState(stateClass, properties);
        if (!AI.isInstance(state)) {
            throw new StateDoesNotImplementException(state, AI);
        }
        return state;
    }

    private void addChildNodeTo(Node<IState<I>> child, Node<IState<I>> parent) {
        parent.addChild(child);
    }

    private void addBranch(StatesTree<I> statesTree, Class<? extends IState<I>> rootClass,
                           Node<IState<I>> addTo,
                           Properties properties) {
        IState<I> newState = constructState(rootClass, properties);
        Node<IState<I>> newNode;
        if (addTo == null) {
            statesTree.setRoot(newState);
            newNode = statesTree.getRoot();
        } else {
            newNode = new Node<IState<I>>(statesTree, addTo, newState);
            addChildNodeTo(newNode, addTo);
        }
        fillBranch(statesTree, newNode, properties);
    }

    private void fillBranch(StatesTree<I> statesTree, Node<IState<I>> node, Properties properties) {
        StateDiscriptor<I> sd = StateDiscriptorFactory.getStateDiscriptor(node.getValue());
        if (sd instanceof ANDStateDiscriptor) {
            Class<? extends IState<I>>[] parts =
                    ((ANDStateDiscriptor<I>)sd).getParts();
            for (Class<? extends IState<I>> part : parts) {
                addBranch(statesTree, part, node, properties);
            }
        } else if (sd instanceof ORStateDiscriptor) {
            addBranch( statesTree,
                    ((ORStateDiscriptor<I>)sd).getEntryPoint(),
                    node,
                    properties);
        }
    }

}
