package ostate.impl;

import ostate.util.properties.Properties;
import ostate.exceptions.InvalidStateException;
import ostate.IStateFactory;
import ostate.IStateMachine;
import ostate.IState;

import java.lang.reflect.InvocationTargetException;

/**
 * Default implementation of {@link ostate.IStateFactory}.
 *
 * Construct states, using constructor
 *
 * <p><code>
 * &lt;State class name&gt;(IStateMachine&lt;I&gt; sm, Properties p, &lt;Other arguments&gt;);
 * </code></p>
 * <p>or</p>
 * <p><code>
 * &lt;State class name&gt;(IStateMachine&lt;I&gt; sm, &lt;Other arguments&gt;);
 * </code></p>
 *
 * @author Pavel Mavrin
 * @version $Id$
 */

public class DefaultStateFactory<I> implements IStateFactory<I> {

    private Object[] args;

    private Class[] argsClasses;

    private IStateMachine<I> sm;

    private final boolean useProperties;

    public DefaultStateFactory(boolean useProperties, Object... args) {
        setArgs(args);
        this.useProperties = useProperties;
    }

    public void setArgs(Object... args) {
        if (useProperties) {
            argsClasses = new Class<?>[args.length + 2];
            argsClasses[0] = IStateMachine.class;
            argsClasses[1] = Properties.class;
            for (int i = 0; i < args.length; i++) {
                argsClasses[i + 2] = args[i].getClass();
            }
            Object[] nargs = new Object[args.length + 2];
            System.arraycopy(args, 0, nargs, 2, args.length);
            this.args = nargs;
        } else {
            argsClasses = new Class<?>[args.length + 1];
            argsClasses[0] = IStateMachine.class;
            for (int i = 0; i < args.length; i++) {
                argsClasses[i + 1] = args[i].getClass();
            }
            Object[] nargs = new Object[args.length + 1];
            System.arraycopy(args, 0, nargs, 1, args.length);
            this.args = nargs;
        }
    }


    public <T extends IState<I>>T constructState(Class<T> stateClass, Properties properties) {
        try {
            if (useProperties) {
                args[1] = properties;
                return stateClass.getConstructor(
                        argsClasses
                ).newInstance(args);
            } else {
                return stateClass.getConstructor(
                        argsClasses
                ).newInstance(args);
            }
        } catch (Exception e) {
            throw new InvalidStateException(
                    "Cannot construct state",
                    stateClass,
                    e);
        }
    }

    public void setStateMachine(IStateMachine<I> sm) {
        this.sm = sm;
        args[0] = sm;
    }
}
